#include <windows.h>

#include <memory.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include "image2d.h"

Image2D::Image2D ( ) {
	w = -1;
	h = -1;
	num_mipmaps = 0;
	bits = NULL;
	texID = -1;
	created = 0;
	min_filter = GL_NEAREST;
	mag_filter = GL_NEAREST;
}

Image2D::~Image2D ( ) {
	if ( bits != NULL )
		delete [] bits;
	bits = NULL;
}

void Image2D::Create ( int width, int height ) {
	// make sure that an image has not been allocated
	Destroy ( );
	
	// allocate the new image and initialize variables.
	bits = new unsigned char[width * height * 4];
	w = width;
	h = height;

	num_mipmaps = 0;
	int x = w > h ? h : w;

	while ( x ) {
		x = x >> 1;
		num_mipmaps++;
	}

	mipmap = new unsigned char *[num_mipmaps];
	for ( int k = 0; k < num_mipmaps; k++ ) {
		mipmap[k] = new unsigned char[(w >> (k+1)) * (h >> (k+1)) * 3];
	}
}

void Image2D::Destroy ( ) {
	if ( bits == NULL )
		delete bits;
	bits = NULL;
}

void Image2D::Clear ( ) {
	// set all the values in the array to zero.
	memset ( bits, 0, w * h * 3 );
}

void Image2D::SetPixel ( int x, int y, int r, int g, int b ) {
	unsigned char *pixels = (unsigned char *)&bits[y * w * 3 + x * 3];
	r = (r > 255 ? 255 : r < 0 ? 0 : r);
	g = (g > 255 ? 255 : g < 0 ? 0 : g);
	b = (b > 255 ? 255 : b < 0 ? 0 : b);
	*pixels++ = r;
	*pixels++ = g;
	*pixels++ = b;
	//((unsigned int)r << 16 | (unsigned int)g << 8 | (unsigned int)b);
}

void Image2D::GetPixel ( int x, int y, int *r, int *g, int *b ) {
	unsigned char *pixels = (unsigned char *)&bits[y * w * 3 + x * 3];
	*r = *pixels++;
	*g = *pixels++;
	*b = *pixels++;
}

void Image2D::Smooth ( ) {
	int x, y, r, g, b;
	int nr, ng, nb;

	for ( y = 0; y < h; y++ ) {
		for ( x = 0; x < w; x++ ) {
			nr = ng = nb = 0;
			GetPixel ( x, y, &r, &g, &b );
			nr += r; ng += g; nb += b;
			GetPixel ( (x + 1)%w, y, &r, &g, &b );
			nr += r; ng += g; nb += b;
			if ( x == 0 )
				GetPixel ( (x - 1) + w, y, &r, &g, &b );
			else
				GetPixel ( (x - 1) % w, y, &r, &g, &b );
			nr += r; ng += g; nb += b;
			GetPixel ( x, (y + 1)%h, &r, &g, &b );
			nr += r; ng += g; nb += b;
			if ( y == 0 )
				GetPixel ( x, (y - 1) + h, &r, &g, &b );
			else
				GetPixel ( x, (y - 1)%h, &r, &g, &b );
			nr += r; ng += g; nb += b;
			nr /= 5;
			ng /= 5;
			nb /= 5;
			SetPixel ( x, y, nr, ng, nb );
		}
	}
}

void Image2D::OpenGLUpload ( ) {
	if ( !created ) {
		glGenTextures ( 1, &texID );
		created = 1;
	}

	glBindTexture ( GL_TEXTURE_2D, texID );

	glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
	glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );

	int i, j, k;
	// create mipmaps
	for ( j = 0; j < h; j++ ) {
		for ( i = 0; i < w; i++ ) {
			for ( k = 0; k < num_mipmaps; k++ ) { 
				mipmap[k][(j >> (k+1)) * 3 * (w >> (k+1)) + (i >> (k+1)) * 3+0] = bits[((j >> (k+1)) << (k+1)) * 3 * w + ((i >> (k+1)) << (k+1)) * 3+0];
				mipmap[k][(j >> (k+1)) * 3 * (w >> (k+1)) + (i >> (k+1)) * 3+1] = bits[((j >> (k+1)) << (k+1)) * 3 * w + ((i >> (k+1)) << (k+1)) * 3+1];
				mipmap[k][(j >> (k+1)) * 3 * (w >> (k+1)) + (i >> (k+1)) * 3+2] = bits[((j >> (k+1)) << (k+1)) * 3 * w + ((i >> (k+1)) << (k+1)) * 3+2];
			}
		}
	}

	glTexImage2D ( GL_TEXTURE_2D, 0, 3, w, h, 0, GL_RGB, GL_UNSIGNED_BYTE, (void *)bits );
	for ( k = 0; k < num_mipmaps; k++ )
		glTexImage2D ( GL_TEXTURE_2D, k+1, 3, w>>(k+1), h>>(k+1), 0, GL_RGB, GL_UNSIGNED_BYTE, (void *)mipmap[k] );
}

void Image2D::GLBind ( ) {
	glBindTexture ( GL_TEXTURE_2D, texID );
	glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, min_filter );
	glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mag_filter );
}